﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Operators;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;
using HeuristicLab.Common;
using HeuristicLab.Parameters;
using HeuristicLab.Data;
using HeuristicLab.Optimization;

namespace HeuristicLab.Problems.MetaOptimization {
  [Item("AlgorithmRunsAnalyzer", "")]
  [StorableClass]
  public class AlgorithmRunsAnalyzer : SingleSuccessorOperator {

    #region Parameter properties
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public LookupParameter<IntValue> GenerationsParameter {
      get { return (LookupParameter<IntValue>)Parameters["Generations"]; }
    }
    public LookupParameter<IntValue> RepetitionsParameter {
      get { return (LookupParameter<IntValue>)Parameters[MetaOptimizationProblem.RepetitionsParameterName]; }
    }
    public ILookupParameter<ParameterConfigurationTree> ParameterConfigurationParameter {
      get { return (ILookupParameter<ParameterConfigurationTree>)Parameters["ParameterConfigurationTree"]; }
    }
    public ILookupParameter<IItemList<IProblem>> ProblemsParameter {
      get { return (ILookupParameter<IItemList<IProblem>>)Parameters[MetaOptimizationProblem.ProblemsParameterName]; }
    }
    public LookupParameter<DoubleArray> ProblemQualityReferencesParameter {
      get { return (LookupParameter<DoubleArray>)Parameters["ProblemQualityReferences"]; }
    }
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public ScopeTreeLookupParameter<IAlgorithm> AlgorithmParameter {
      get { return (ScopeTreeLookupParameter<IAlgorithm>)Parameters["Algorithm"]; }
    }
    public ScopeTreeLookupParameter<IntValue> ProblemIndexParameter {
      get { return (ScopeTreeLookupParameter<IntValue>)Parameters["ProblemIndex"]; }
    }
    public ScopeTreeLookupParameter<IntValue> RepetitionIndexParameter {
      get { return (ScopeTreeLookupParameter<IntValue>)Parameters["RepetitionIndex"]; }
    }
    #endregion

    [StorableConstructor]
    protected AlgorithmRunsAnalyzer(bool deserializing) : base(deserializing) { }
    public AlgorithmRunsAnalyzer() : base() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The pseudo random number generator which should be used to initialize the new random permutation."));
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The evaluated quality of the ParameterVector."));
      Parameters.Add(new LookupParameter<IntValue>("Generations", ""));
      Parameters.Add(new LookupParameter<IntValue>(MetaOptimizationProblem.RepetitionsParameterName, "Number of evaluations on one problem."));
      Parameters.Add(new LookupParameter<ParameterConfigurationTree>("ParameterConfigurationTree", ""));
      Parameters.Add(new LookupParameter<IItemList<IProblem>>(MetaOptimizationProblem.ProblemsParameterName, ""));
      Parameters.Add(new LookupParameter<DoubleArray>("ProblemQualityReferences", ""));
      Parameters.Add(new LookupParameter<ResultCollection>("Results", ""));
      Parameters.Add(new ScopeTreeLookupParameter<IAlgorithm>("Algorithm", "The finished algorithms containing Runs."));
      Parameters.Add(new ScopeTreeLookupParameter<IntValue>("ProblemIndex", "The index of the problem an algorithm was executed with."));
      Parameters.Add(new ScopeTreeLookupParameter<IntValue>("RepetitionIndex", "The index of the repetition"));
    }
    protected AlgorithmRunsAnalyzer(AlgorithmRunsAnalyzer original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new AlgorithmRunsAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      ParameterConfigurationTree parameterConfiguration = ParameterConfigurationParameter.ActualValue;
      ItemArray<IAlgorithm> algorithms = AlgorithmParameter.ActualValue;
      ItemArray<IntValue> problemIndices = ProblemIndexParameter.ActualValue;
      ItemArray<IntValue> repetitionIndices = RepetitionIndexParameter.ActualValue;
      IEnumerable<string> parameterNames = parameterConfiguration.GetOptimizedParameterNames();
      IItemList<IProblem> problems = ProblemsParameter.ActualValue;
      int repetitions = RepetitionsParameter.ActualValue.Value;
      var resultNames = new List<string> { "BestQuality", "Execution Time" };
      int currentGeneration = GenerationsParameter.ActualValue != null ? GenerationsParameter.ActualValue.Value : 0;
      double[] referenceQualities = GetReferenceQualities(problems.Count);
      ResultCollection results = ResultsParameter.ActualValue;

      if (algorithms.All(x => x.Runs.Count == 1)) {
        var runs = new RunCollection();
        var qualities = new double[problems.Count][];
        var executionTimes = new TimeSpan[problems.Count][];

        for (int i = 0; i < problems.Count; i++) {
          qualities[i] = new double[repetitions];
          executionTimes[i] = new TimeSpan[repetitions];
        }

        for (int i = 0; i < algorithms.Length; i++) {
          int problemIndex = problemIndices[i].Value;
          int repetitionIndex = repetitionIndices[i].Value;

          IRun run = algorithms[i].Runs.First();
          MetaOptimizationUtil.ClearResults(run, resultNames);
          MetaOptimizationUtil.ClearParameters(run, parameterNames);
          run.Results.Add("Meta.FromCache", new BoolValue(false));
          run.Results.Add("Meta.Generation", new IntValue(currentGeneration));
          run.Results.Add("Meta.ProblemIndex", new IntValue(problemIndex));
          run.Name = string.Format("{0} Problem {1} Run {2}", parameterConfiguration.ParameterInfoString, problemIndex, repetitionIndex);
          qualities[problemIndex][repetitionIndex] = (((DoubleValue)run.Results["BestQuality"]).Value);
          executionTimes[problemIndex][repetitionIndex] = (((TimeSpanValue)run.Results["Execution Time"]).Value);
          runs.Add((IRun)run.Clone());
        }

        parameterConfiguration.AverageExecutionTimes = new ItemList<TimeSpanValue>(executionTimes.Select(t => new TimeSpanValue(TimeSpan.FromMilliseconds(t.Average(ts => ts.TotalMilliseconds)))));
        parameterConfiguration.Repetitions = new IntValue(repetitions);
        parameterConfiguration.BestQualities = new DoubleArray(qualities.Select(q => q.Min()).ToArray()); // todo: respect Maximization:true/false
        parameterConfiguration.AverageQualities = new DoubleArray(qualities.Select(q => q.Average()).ToArray());
        parameterConfiguration.WorstQualities = new DoubleArray(qualities.Select(q => q.Max()).ToArray()); // todo: respect Maximization:true/false
        parameterConfiguration.QualityVariances = new DoubleArray(qualities.Select(q => q.Variance()).ToArray());
        parameterConfiguration.QualityStandardDeviations = new DoubleArray(qualities.Select(q => q.StandardDeviation()).ToArray());
        parameterConfiguration.Runs = runs;

        this.QualityParameter.ActualValue = new DoubleValue(MetaOptimizationUtil.NormalizeQualities(parameterConfiguration, referenceQualities));
      } else {
        // something terrible happened -> most probably due to invalid parameters.
        // penalty with worst quality from latest generation!

        double penaltyValue = results.ContainsKey("CurrentWorstQuality") ? ((DoubleValue)results["CurrentWorstQuality"]).Value : referenceQualities.Max(); // todo: respect min/max
        this.QualityParameter.ActualValue = new DoubleValue(penaltyValue);
      }

      return base.Apply();
    }

    private double[] GetReferenceQualities(int problemsCount) {
      double[] referenceQualities;
      if (ProblemQualityReferencesParameter.ActualValue == null) {
        // this is generation zero. no reference qualities for normalization have been calculated yet. in this special case the ReferenceQualityAnalyzer will do the normalization
        referenceQualities = new double[problemsCount];
        for (int i = 0; i < referenceQualities.Length; i++) {
          referenceQualities[i] = 1;
        }
      } else {
        referenceQualities = ProblemQualityReferencesParameter.ActualValue.ToArray();
      }
      return referenceQualities;
    }
  }
}
